//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

// Pawcio: BC Download
#include "stdafx.h"
#include "emule.h"
#include "DownloadBandwidthThrottler.h"
#include "preferences.h"
#include "updownclient.h"
#include "OtherFunctions.h"
#include "Listensocket.h"
#include "PartFile.h" // Pawcio for enkeyDev: ICS

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/**
* The constructor starts the thread.
*/
DownloadBandwidthThrottler::DownloadBandwidthThrottler(void) {

	threadEndedEvent = new CEvent(0, 1);
	pauseEvent = new CEvent(TRUE, TRUE);

	doRun = true;
	timeDown = thePrefs.GetBCTimeDown();
	AfxBeginThread(RunProc, (LPVOID)this);
}

/**
* The destructor stops the thread. If the thread has already stoppped, destructor does nothing.
*/
DownloadBandwidthThrottler::~DownloadBandwidthThrottler(void) {
	EndThread();
	delete threadEndedEvent;
	delete pauseEvent;
}

void DownloadBandwidthThrottler::AddToStandardList(CClientReqSocket* socket) {
	if (socket) {
		receiveLocker.Lock();
		AddToStandardListNoLock(socket);
		receiveLocker.Unlock();
	}
}

void DownloadBandwidthThrottler::AddToStandardListNoLock(CClientReqSocket* socket) {
	if (socket && !socket->onDownQueue){
		downloadinglist.AddTail(socket);
		socket->onDownQueue = true;
	}
}

void DownloadBandwidthThrottler::RemoveFromStandardList(CClientReqSocket* socket) {
	if (socket){
		receiveLocker.Lock();

		RemoveFromStandardListNoLock(socket);

		receiveLocker.Unlock();
	}
}

void DownloadBandwidthThrottler::RemoveFromStandardListNoLock(CClientReqSocket* socket) {
	if (socket && socket->onDownQueue) {
		// Find the slot
		POSITION pos = downloadinglist.Find(socket);
		if (pos){
			if (pos == receivingPos)
				downloadinglist.GetNext(receivingPos);
			socket->onDownQueue = false;
			downloadinglist.RemoveAt(pos);
		}
	}
}

bool DownloadBandwidthThrottler::QueueForReceivingPacket(ThrottledControlSocket* socket) {
	if (socket){
		// Get critical section
		receiveLocker.Lock();
		if(doRun && socket && !socket->onDownControlQueue && !socket->onDownQueue) {
			socket->onDownControlQueue = true;
			m_ControlQueue_list.AddTail(socket);
		}

		bool ret = socket->onDownControlQueue || socket->onDownQueue;
		// End critical section
		receiveLocker.Unlock();
		return ret;
	}
	return false;
}

void DownloadBandwidthThrottler::RemoveFromAllQueues(ThrottledControlSocket* socket, bool lock) {
	if (socket){
		if (socket->onDownControlQueue){
			if(lock) {
				// Get critical section
				receiveLocker.Lock();
			}

			if(doRun) {
				// Remove this socket from control packet queue
				POSITION pos = m_ControlQueue_list.Find(socket);
				if (pos){
					socket->onDownControlQueue = false;
					m_ControlQueue_list.RemoveAt(pos);
				}
			}
			if(lock) {
				// End critical section
				receiveLocker.Unlock();
			}
		}
	}
}

void DownloadBandwidthThrottler::RemoveFromAllQueues(ThrottledControlSocket* socket) {
	if (socket){
		// Get critical section
		receiveLocker.Lock();

		if(doRun) {
			RemoveFromAllQueues(socket, false);
		}

		// End critical section
		receiveLocker.Unlock();
	}
}

/**
* Make the thread exit. This method will not return until the thread has stopped
* looping. This guarantees that the thread will not access the CEMSockets after this
* call has exited.
*/
void DownloadBandwidthThrottler::EndThread() {
	receiveLocker.Lock();

	// signal the thread to stop looping and exit.
	doRun = false;

	receiveLocker.Unlock();

	Pause(false);

	// wait for the thread to signal that it has stopped looping.
	threadEndedEvent->Lock();
}

void DownloadBandwidthThrottler::Pause(bool paused) {
	if(paused) {
		pauseEvent->ResetEvent();
	} else {
		pauseEvent->SetEvent();
	}
}

/**
* Start the thread. Called from the constructor in this class.
*
* @param pParam
*
* @return
*/
UINT AFX_CDECL DownloadBandwidthThrottler::RunProc(LPVOID pParam) {
	DbgSetThreadName("DownloadBandwidthThrottler");
	InitThreadLocale();
	DownloadBandwidthThrottler* downloadBandwidthThrottler = (DownloadBandwidthThrottler*)pParam;

	return downloadBandwidthThrottler->RunInternal();
}

/**
* @return always returns 0.
*/
UINT DownloadBandwidthThrottler::RunInternal() {
	DWORD lastLoopTick = ::GetTickCount();
	float allowedDataRate = 0;

	receivingPos = 0;
	toReceive = 0;
	while(doRun) {
		pauseEvent->Lock();

		DWORD timeSinceLastLoop = ::GetTickCount() - lastLoopTick;

		// Get current speed from UploadSpeedSense
		allowedDataRate = theApp.BC->GetMaxDownload();

		uint32 minFragSize = thePrefs.GetMSS();
		if (allowedDataRate < 5.0)
			minFragSize = 536;
		else if (allowedDataRate >= 24.0)
			minFragSize *= 2;

#define TIME_BETWEEN_DOWNLOAD_LOOPS timeDown
		// sleep for just as long as we need to get back to having minFragSize/2 bytes to send
		uint32 sleepTime = max((minFragSize/2)/allowedDataRate, TIME_BETWEEN_DOWNLOAD_LOOPS);

		uint32 toSleep = TIME_BETWEEN_DOWNLOAD_LOOPS;
		if(timeSinceLastLoop < sleepTime)
			toSleep = sleepTime - timeSinceLastLoop;
		if (toSleep < 1)
			toSleep = 1;
		else if (toSleep > 250)
			toSleep = 250;
		Sleep(toSleep);

		const DWORD thisLoopTick = ::GetTickCount();
		timeSinceLastLoop = thisLoopTick - lastLoopTick;
		lastLoopTick = thisLoopTick;

		if (allowedDataRate == UNLIMITED)
			toReceive = 0x7fffffff; //max signed value;
		else {
			uint32 currLoopAddOn = allowedDataRate * 1.024f * timeSinceLastLoop;

			if (toReceive > 0 && toReceive + (int) currLoopAddOn < 0)
				toReceive = 0;
			toReceive += currLoopAddOn;

			if (toReceive > allowedDataRate * 1.024f * 1000) {
				toReceive = allowedDataRate * 1.024f * 1000;
			}
		}

		receiveLocker.Lock();


		//Serve control packets
		int tempToReceive;
		POSITION prev_pos = NULL;
		POSITION controlPos = m_ControlQueue_list.GetHeadPosition();
		while (controlPos){
			tempToReceive = toReceive;
			prev_pos = controlPos;
			ThrottledControlSocket* cur_socket = m_ControlQueue_list.GetNext(controlPos);
			if (cur_socket->onDownQueue){
				cur_socket->onDownControlQueue = false;
				m_ControlQueue_list.RemoveAt(prev_pos);
				continue;
			}
			if (cur_socket->IsPriorityReceive() || !thePrefs.IsIncludeOverhead())
				cur_socket->Receive(0x7fffffff);
			else if (tempToReceive >= (int) minFragSize/2)
				cur_socket->Receive(tempToReceive);
		}

		uint32 notFullReceive = downloadinglist.GetCount();
		while ((tempToReceive = toReceive) >= (int) minFragSize && notFullReceive){
			if (!receivingPos)
				receivingPos=downloadinglist.GetHeadPosition();
			CClientReqSocket* cur_socket = downloadinglist.GetNext(receivingPos);
			cur_socket->Receive(tempToReceive);
			notFullReceive--;
		}
		receiveLocker.Unlock();
	}

	threadEndedEvent->SetEvent();

	receiveLocker.Lock();

	m_ControlQueue_list.RemoveAll();
	downloadinglist.RemoveAll();

	receiveLocker.Unlock();

	return 0;
}

void DownloadBandwidthThrottler::ReSortDownloadSlots(CUpDownClient* client){
	if (client){
		receiveLocker.Lock();
		// Remove the found Client from UploadBandwidthThrottler
		theApp.downloadBandwidthThrottler->RemoveFromStandardListNoLock(client->socket);
		theApp.downloadBandwidthThrottler->RemoveFromStandardListNoLock((CClientReqSocket*) client->m_pPCDownSocket);
		theApp.downloadBandwidthThrottler->RemoveFromStandardListNoLock((CClientReqSocket*) client->m_pWCDownSocket);
		theApp.downloadBandwidthThrottler->AddToStandardListNoLock(client->GetFileDownloadSocket());
		receiveLocker.Unlock();
	}
}